//! Directory synchronized handshake, Request parsing

use crate::{
    auth,
    connection::{
        handshake::{
            self,
            dirsync::{DirSync, Nonce},
            Error,
        },
        ProtocolVersion, ID,
    },
    enc::{
        asym::{ExchangePubKey, KeyExchangeKind, KeyID},
        hkdf,
        sym::{self, NonceLen, TagLen},
    },
};

/// Client request of a directory synchronized handshake
#[derive(Debug, Clone, PartialEq)]
pub struct Req {
    /// Id of the server key used for the key exchange
    pub key_id: KeyID,
    /// Selected key exchange
    pub exchange: KeyExchangeKind,
    /// Selected hkdf
    pub hkdf: hkdf::Kind,
    /// Selected cipher
    pub cipher: sym::Kind,
    /// Client ephemeral public key used for key exchanges
    pub exchange_key: ExchangePubKey,
    /// encrypted data
    pub data: State,
    // SECURITY: TODO: Add padding to min: 1200 bytes
    //   to avoid amplification attaks
    //   also: 1200 < 1280 to allow better vpn compatibility
}

impl Req {
    /// return the offset of the encrypted data
    /// NOTE: starts from the beginning of the fenrir packet
    pub fn encrypted_offset(&self) -> usize {
        ProtocolVersion::len()
            + handshake::ID::len()
            + KeyID::len()
            + KeyExchangeKind::len()
            + hkdf::Kind::len()
            + sym::Kind::len()
            + self.exchange_key.kind().pub_len()
    }
    /// return the total length of the cleartext data
    pub fn encrypted_length(
        &self,
        head_len: NonceLen,
        tag_len: TagLen,
    ) -> usize {
        match &self.data {
            State::ClearText(data) => data.len() + head_len.0 + tag_len.0,
            State::CipherText(length) => *length,
        }
    }
    /// actual length of the directory synchronized request
    pub fn len(&self) -> usize {
        KeyID::len()
            + KeyExchangeKind::len()
            + hkdf::Kind::len()
            + sym::Kind::len()
            + self.exchange_key.kind().pub_len()
            + self.cipher.nonce_len().0
            + self.data.len()
            + self.cipher.tag_len().0
    }
    /// Serialize into raw bytes
    /// NOTE: assumes that there is exactly as much buffer as needed
    pub fn serialize(
        &self,
        head_len: NonceLen,
        tag_len: TagLen,
        out: &mut [u8],
    ) {
        out[0..2].copy_from_slice(&self.key_id.0.to_le_bytes());
        out[2] = self.exchange as u8;
        out[3] = self.hkdf as u8;
        out[4] = self.cipher as u8;
        let key_len = self.exchange_key.len();
        let written_next = 5 + key_len;
        self.exchange_key.serialize_into(&mut out[5..written_next]);
        let written = written_next;
        if let State::ClearText(data) = &self.data {
            let from = written + head_len.0;
            let to = out.len() - tag_len.0;
            data.serialize(&mut out[from..to]);
        } else {
            unreachable!();
        }
    }
}

impl handshake::Parsing for Req {
    fn deserialize(raw: &[u8]) -> Result<handshake::Data, Error> {
        const MIN_PKT_LEN: usize = 10;
        if raw.len() < MIN_PKT_LEN {
            return Err(Error::NotEnoughData);
        }
        let key_id: KeyID =
            KeyID(u16::from_le_bytes(raw[0..2].try_into().unwrap()));
        use ::num_traits::FromPrimitive;
        let exchange: KeyExchangeKind = match KeyExchangeKind::from_u8(raw[2]) {
            Some(exchange) => exchange,
            None => return Err(Error::Parsing),
        };
        let hkdf: hkdf::Kind = match hkdf::Kind::from_u8(raw[3]) {
            Some(exchange) => exchange,
            None => return Err(Error::Parsing),
        };
        let cipher: sym::Kind = match sym::Kind::from_u8(raw[4]) {
            Some(cipher) => cipher,
            None => return Err(Error::Parsing),
        };
        const CURR_SIZE: usize = KeyID::len()
            + KeyExchangeKind::len()
            + hkdf::Kind::len()
            + sym::Kind::len();
        let (exchange_key, len) =
            match ExchangePubKey::deserialize(&raw[CURR_SIZE..]) {
                Ok(exchange_key) => exchange_key,
                Err(e) => return Err(e.into()),
            };
        let data = State::CipherText(raw.len() - (CURR_SIZE + len));
        Ok(handshake::Data::DirSync(DirSync::Req(Self {
            key_id,
            exchange,
            hkdf,
            cipher,
            exchange_key,
            data,
        })))
    }
}

/// Quick way to avoid mixing cipher and clear text
#[derive(Debug, Clone, PartialEq)]
pub enum State {
    /// Data is still encrytped, we only keep the length
    CipherText(usize),
    /// Client data, decrypted and parsed
    ClearText(Data),
}
impl State {
    /// The length of the data
    pub fn len(&self) -> usize {
        match self {
            State::CipherText(len) => *len,
            State::ClearText(data) => data.len(),
        }
    }
    /// parse the cleartext
    pub fn deserialize_as_cleartext(
        &mut self,
        raw: &[u8],
    ) -> Result<(), Error> {
        let clear = match self {
            State::CipherText(len) => {
                assert!(
                    *len > raw.len(),
                    "DirSync::State::CipherText length mismatch"
                );
                match Data::deserialize(raw) {
                    Ok(clear) => clear,
                    Err(e) => return Err(e),
                }
            }
            _ => return Err(Error::Parsing),
        };
        *self = State::ClearText(clear);
        Ok(())
    }
}

/// Informations needed for authentication
#[derive(Debug, Clone, PartialEq)]
pub struct AuthInfo {
    /// User of the domain
    pub user: auth::UserID,
    /// Authentication token
    pub token: auth::Token,
    /// service ID that we want to use on the domain
    pub service_id: auth::ServiceID,
    /// subdomain on which we authenticate
    // SECURITY: TODO: this should be padded to multiples of 32 bytes or so
    // reason: packet length can let you infer the authentication domain
    pub domain: auth::Domain,
}

impl AuthInfo {
    /// Minimum length of the authentication info
    pub const MIN_PKT_LEN: usize =
        auth::UserID::len() + auth::Token::len() + auth::ServiceID::len() + 1;
    /// Actual length of the authentication info
    pub fn len(&self) -> usize {
        Self::MIN_PKT_LEN + self.domain.len()
    }
    /// serialize into a buffer
    /// Note: assumes there is enough space
    pub fn serialize(&self, out: &mut [u8]) {
        out[..auth::UserID::len()].copy_from_slice(&self.user.0);
        const WRITTEN_TOKEN: usize = auth::UserID::len() + auth::Token::len();
        out[auth::UserID::len()..WRITTEN_TOKEN].copy_from_slice(&self.token.0);
        const WRITTEN_SERVICE_ID: usize =
            WRITTEN_TOKEN + auth::ServiceID::len();
        out[WRITTEN_TOKEN..WRITTEN_SERVICE_ID]
            .copy_from_slice(&self.service_id.0);
        let domain_len = self.domain.0.as_bytes().len() as u8;
        out[WRITTEN_SERVICE_ID] = domain_len;
        const WRITTEN_DOMAIN_LEN: usize = WRITTEN_SERVICE_ID + 1;
        out[WRITTEN_DOMAIN_LEN..].copy_from_slice(&self.domain.0.as_bytes());
    }
    /// deserialize from raw bytes
    pub fn deserialize(raw: &[u8]) -> Result<Self, Error> {
        if raw.len() < Self::MIN_PKT_LEN {
            return Err(Error::NotEnoughData);
        }
        let raw_user_id: [u8; auth::UserID::len()];
        let raw_token: [u8; auth::Token::len()];
        let raw_service_id: [u8; auth::ServiceID::len()];
        let mut start = 0;
        let mut end = auth::UserID::len();
        raw_user_id = raw[start..end].try_into().unwrap();
        let user: auth::UserID = raw_user_id.into();
        start = end;
        end = end + auth::Token::len();
        raw_token = raw[start..end].try_into().unwrap();
        let token: auth::Token = raw_token.into();
        start = end;
        end = end + auth::ServiceID::len();
        raw_service_id = raw[start..end].try_into().unwrap();
        let service_id: auth::ServiceID = raw_service_id.into();
        start = end;
        let domain_len = raw[start];
        start = start + 1;
        end = start + domain_len as usize;
        if raw.len() < end {
            return Err(Error::NotEnoughData);
        }
        let domain: auth::Domain = match raw[start..end].try_into() {
            Ok(domain) => domain,
            Err(_) => return Err(Error::Parsing),
        };
        Ok(Self {
            user,
            token,
            service_id,
            domain,
        })
    }
}

/// Decrypted request data
#[derive(Debug, Clone, PartialEq)]
pub struct Data {
    /// Random nonce, the client can use this to track multiple key exchanges
    pub nonce: Nonce,
    /// Client key id so the client can use and rotate keys
    pub client_key_id: KeyID,
    /// Receiving connection id for the client
    pub id: ID,
    /// Authentication data
    pub auth: AuthInfo,
}
impl Data {
    /// actual length of the request data
    pub fn len(&self) -> usize {
        Nonce::len() + KeyID::len() + ID::len() + self.auth.len()
    }
    /// Minimum byte length of the request data
    pub const MIN_PKT_LEN: usize =
        16 + KeyID::len() + ID::len() + AuthInfo::MIN_PKT_LEN;
    /// serialize into a buffer
    /// Note: assumes there is enough space
    pub fn serialize(&self, out: &mut [u8]) {
        out[..Nonce::len()].copy_from_slice(&self.nonce.0);
        const WRITTEN_KEY: usize = Nonce::len() + KeyID::len();
        out[Nonce::len()..WRITTEN_KEY]
            .copy_from_slice(&self.client_key_id.0.to_le_bytes());
        const WRITTEN: usize = WRITTEN_KEY;
        const WRITTEN_ID: usize = WRITTEN + 8;
        out[WRITTEN..WRITTEN_ID]
            .copy_from_slice(&self.id.as_u64().to_le_bytes());
        self.auth.serialize(&mut out[WRITTEN_ID..]);
    }
    /// Parse the cleartext raw data
    pub fn deserialize(raw: &[u8]) -> Result<Self, Error> {
        if raw.len() < Self::MIN_PKT_LEN {
            return Err(Error::NotEnoughData);
        }
        let mut start = 0;
        let mut end = 16;
        let raw_sized: &[u8; 16] = raw[start..end].try_into().unwrap();
        let nonce: Nonce = raw_sized.into();
        start = end;
        end = end + KeyID::len();
        let client_key_id =
            KeyID(u16::from_le_bytes(raw[start..end].try_into().unwrap()));
        start = end;
        end = end + ID::len();
        let id: ID =
            u64::from_le_bytes(raw[start..end].try_into().unwrap()).into();
        if id.is_handshake() {
            return Err(Error::Parsing);
        }
        start = end;
        //end = raw.len();
        let auth = AuthInfo::deserialize(&raw[start..])?;
        Ok(Self {
            nonce,
            client_key_id,
            id,
            auth,
        })
    }
}
