//! Directory synchronized handshake, Response parsing

use crate::{
    connection::{
        handshake::{
            self,
            dirsync::{DirSync, Nonce},
            Error,
        },
        ProtocolVersion, ID,
    },
    enc::{
        asym::KeyID,
        sym::{NonceLen, TagLen},
        Secret,
    },
};

/// Server response in a directory synchronized handshake
#[derive(Debug, Clone, PartialEq)]
pub struct Resp {
    /// Tells the client with which key the exchange was done
    pub client_key_id: KeyID,
    /// actual response data, might be encrypted
    pub data: State,
}

impl handshake::Parsing for Resp {
    fn deserialize(raw: &[u8]) -> Result<handshake::Data, Error> {
        const MIN_PKT_LEN: usize = 68;
        if raw.len() < MIN_PKT_LEN {
            return Err(Error::NotEnoughData);
        }
        let client_key_id: KeyID =
            KeyID(u16::from_le_bytes(raw[0..KeyID::len()].try_into().unwrap()));
        Ok(handshake::Data::DirSync(DirSync::Resp(Self {
            client_key_id,
            data: State::CipherText(raw[KeyID::len()..].len()),
        })))
    }
}

impl Resp {
    /// return the offset of the encrypted data
    /// NOTE: starts from the beginning of the fenrir packet
    pub fn encrypted_offset(&self) -> usize {
        ProtocolVersion::len() + handshake::ID::len() + KeyID::len()
    }
    /// return the total length of the cleartext data
    pub fn encrypted_length(
        &self,
        head_len: NonceLen,
        tag_len: TagLen,
    ) -> usize {
        match &self.data {
            State::ClearText(_data) => Data::len() + head_len.0 + tag_len.0,
            State::CipherText(len) => *len,
        }
    }
    /// Total length of the response handshake
    pub fn len(&self, head_len: NonceLen, tag_len: TagLen) -> usize {
        KeyID::len() + head_len.0 + self.data.len() + tag_len.0
    }
    /// Serialize into raw bytes
    /// NOTE: assumes that there is exactly as much buffer as needed
    pub fn serialize(
        &self,
        head_len: NonceLen,
        _tag_len: TagLen,
        out: &mut [u8],
    ) {
        out[0..KeyID::len()]
            .copy_from_slice(&self.client_key_id.0.to_le_bytes());
        let start_data = KeyID::len() + head_len.0;
        let end_data = start_data + self.data.len();
        self.data.serialize(&mut out[start_data..end_data]);
    }
}

/// Quick way to avoid mixing cipher and clear text
#[derive(Debug, Clone, PartialEq)]
pub enum State {
    /// Server data, still in ciphertext
    CipherText(usize),
    /// Parsed, cleartext server data
    ClearText(Data),
}
impl State {
    /// The length of the data
    pub fn len(&self) -> usize {
        match self {
            State::CipherText(len) => *len,
            State::ClearText(_) => Data::len(),
        }
    }
    /// parse the cleartext
    pub fn deserialize_as_cleartext(
        &mut self,
        raw: &[u8],
    ) -> Result<(), Error> {
        let clear = match self {
            State::CipherText(len) => {
                assert!(
                    *len > raw.len(),
                    "DirSync::State::CipherText length mismatch"
                );
                match Data::deserialize(raw) {
                    Ok(clear) => clear,
                    Err(e) => return Err(e),
                }
            }
            _ => return Err(Error::Parsing),
        };
        *self = State::ClearText(clear);
        Ok(())
    }
    /// Serialize the still cleartext data
    pub fn serialize(&self, out: &mut [u8]) {
        if let State::ClearText(clear) = &self {
            clear.serialize(out);
        }
    }
}

/// Decrypted response data
#[derive(Debug, Clone, PartialEq)]
pub struct Data {
    /// Client nonce, copied from the request
    pub client_nonce: Nonce,
    /// Server Connection ID
    pub id: ID,
    /// Service Connection ID
    pub service_connection_id: ID,
    /// Service encryption key
    pub service_key: Secret,
}

impl Data {
    /// Return the expected length for buffer allocation
    pub fn len() -> usize {
        Nonce::len() + ID::len() + ID::len() + Secret::len()
    }
    /// Serialize the data into a buffer
    /// NOTE: assumes that there is exactly asa much buffer as needed
    pub fn serialize(&self, out: &mut [u8]) {
        let mut start = 0;
        let mut end = Nonce::len();
        out[start..end].copy_from_slice(&self.client_nonce.0);
        start = end;
        end = end + ID::len();
        self.id.serialize(&mut out[start..end]);
        start = end;
        end = end + ID::len();
        self.service_connection_id.serialize(&mut out[start..end]);
        start = end;
        end = end + Secret::len();
        out[start..end].copy_from_slice(self.service_key.as_ref());
    }
    /// Parse the cleartext raw data
    pub fn deserialize(raw: &[u8]) -> Result<Self, Error> {
        let raw_sized: &[u8; 16] = raw[..Nonce::len()].try_into().unwrap();
        let client_nonce: Nonce = raw_sized.into();
        let end = Nonce::len() + ID::len();
        let id: ID =
            u64::from_le_bytes(raw[Nonce::len()..end].try_into().unwrap())
                .into();
        if id.is_handshake() {
            return Err(Error::Parsing);
        }
        let parsed = end;
        let end = parsed + ID::len();
        let service_connection_id: ID =
            u64::from_le_bytes(raw[parsed..end].try_into().unwrap()).into();
        if service_connection_id.is_handshake() {
            return Err(Error::Parsing);
        }
        let parsed = end;
        let end = parsed + Secret::len();
        let raw_secret: &[u8; 32] = raw[parsed..end].try_into().unwrap();
        let service_key = raw_secret.into();

        Ok(Self {
            client_nonce,
            id,
            service_connection_id,
            service_key,
        })
    }
}
