//
//! Raw packet handling, encryption, decryption, parsing

use crate::{
    connection,
    enc::sym::{NonceLen, TagLen},
};

/// Enumerate the possible data in a fenrir packet
#[derive(Debug, Clone)]
pub enum Data {
    /// A parsed handshake packet
    Handshake(super::Handshake),
    /// Raw packet. we only have the connection ID and packet length
    Raw(usize),
}

impl Data {
    /// total length of the data in bytes
    pub fn len(&self, head_len: NonceLen, tag_len: TagLen) -> usize {
        match self {
            Data::Handshake(h) => h.len(head_len, tag_len),
            Data::Raw(len) => *len,
        }
    }
    /// serialize data into bytes
    /// NOTE: assumes that there is exactly asa much buffer as needed
    pub fn serialize(
        &self,
        head_len: NonceLen,
        tag_len: TagLen,
        out: &mut [u8],
    ) {
        assert!(
            self.len(head_len, tag_len) == out.len(),
            "Data: wrong buffer length"
        );
        match self {
            Data::Handshake(h) => h.serialize(head_len, tag_len, out),
            Data::Raw(_) => {
                ::tracing::error!("Tried to serialize a raw Data!");
            }
        }
    }
}

const MIN_PACKET_BYTES: usize = 16;

/// Fenrir packet structure
#[derive(Debug, Clone)]
pub struct Packet {
    /// Id of the Fenrir connection.
    pub id: connection::ID,
    /// actual data inside the packet
    pub data: Data,
}

impl Packet {
    /// New recevied packet, yet unparsed
    pub fn deserialize_id(raw: &[u8]) -> Result<Self, ()> {
        // TODO: proper min_packet length. 16 is too conservative.
        if raw.len() < MIN_PACKET_BYTES {
            return Err(());
        }
        let raw_id: [u8; 8] = (raw[..8]).try_into().expect("unreachable");
        Ok(Packet {
            id: raw_id.into(),
            data: Data::Raw(raw.len()),
        })
    }
    /// get the total length of the packet
    pub fn len(&self, head_len: NonceLen, tag_len: TagLen) -> usize {
        connection::ID::len() + self.data.len(head_len, tag_len)
    }
    /// serialize packet into buffer
    /// NOTE: assumes that there is exactly asa much buffer as needed
    pub fn serialize(
        &self,
        head_len: NonceLen,
        tag_len: TagLen,
        out: &mut [u8],
    ) {
        assert!(
            out.len() > connection::ID::len(),
            "Packet: not enough buffer to serialize"
        );
        self.id.serialize(&mut out[0..connection::ID::len()]);
        self.data.serialize(
            head_len,
            tag_len,
            &mut out[connection::ID::len()..],
        );
    }
}
