use crate::{
    connection::stream::{self, Chunk, uud::*},
    enc::Random,
};

#[::tracing_test::traced_test]
#[test]
fn test_stream_uud_sequential() {
    let rand = Random::new();
    let mut uud = UnreliableUnorderedDatagram::with_window_size(&rand, 1048576);

    let mut data = Vec::with_capacity(1024);
    data.resize(data.capacity(), 0);
    rand.fill(&mut data[..]);

    let start = uud.window_start.0;

    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: true,
        flag_end: false,
        sequence: start,
        data: &data[..512],
    };
    let got = uud.get();
    assert!(&got[..] == &[], "uud: got data?");
    let _ = uud.recv(chunk);
    let got = uud.get();
    assert!(
        &data[..512] == &got[..],
        "UUD1: DIFF: {:?} {:?}",
        &data[..512].len(),
        &got[..].len()
    );
    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: false,
        flag_end: true,
        sequence: start + 512,
        data: &data[512..],
    };
    let _ = uud.recv(chunk);
    let got = uud.get();
    assert!(
        &data[512..] == &got[..],
        "UUD2: DIFF: {:?} {:?}",
        &data[512..].len(),
        &got[..].len()
    );
}

#[::tracing_test::traced_test]
#[test]
fn test_stream_uud_retransmit() {
    let rand = Random::new();
    let max_window: usize = 100;
    let mut uud =
        UnreliableUnorderedDatagram::with_window_size(&rand, max_window as u32);

    let mut data = Vec::with_capacity(120);
    data.resize(data.capacity(), 0);
    for i in 0..data.len() {
        data[i] = i as u8;
    }

    let start = uud.window_start.0;

    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: true,
        flag_end: false,
        sequence: start,
        data: &data[..40],
    };
    let _ = uud.recv(chunk);
    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: false,
        flag_end: false,
        sequence: start + 50,
        data: &data[50..60],
    };
    let _ = uud.recv(chunk);
    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: false,
        flag_end: false,
        sequence: start + 40,
        data: &data[40..60],
    };
    let _ = uud.recv(chunk);
    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: false,
        flag_end: false,
        sequence: start + 80,
        data: &data[80..],
    };
    let _ = uud.recv(chunk);
    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: false,
        flag_end: false,
        sequence: start + 50,
        data: &data[50..90],
    };
    let _ = uud.recv(chunk);
    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: false,
        flag_end: false,
        sequence: start + (max_window as u32),
        data: &data[max_window..],
    };
    let _ = uud.recv(chunk);
    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: false,
        flag_end: true,
        sequence: start + 90,
        data: &data[90..max_window],
    };
    let _ = uud.recv(chunk);
    let got = uud.get();
    assert!(
        &data[..max_window] == &got[..],
        "DIFF:\n {:?}\n {:?}",
        &data[..max_window],
        &got[..],
    );
}
#[::tracing_test::traced_test]
#[test]
fn test_stream_uud_rolling() {
    let rand = Random::new();
    let max_window: usize = 100;
    let mut uud =
        UnreliableUnorderedDatagram::with_window_size(&rand, max_window as u32);

    let mut data = Vec::with_capacity(120);
    data.resize(data.capacity(), 0);
    for i in 0..data.len() {
        data[i] = i as u8;
    }

    let start = uud.window_start.0;

    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: true,
        flag_end: false,
        sequence: start,
        data: &data[..40],
    };
    let _ = uud.recv(chunk);
    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: true,
        flag_end: false,
        sequence: start + 50,
        data: &data[50..100],
    };
    let _ = uud.recv(chunk);
    let got = uud.get();
    assert!(
        &data[..40] == &got[..],
        "DIFF:\n {:?}\n {:?}",
        &data[..40],
        &got[..],
    );
    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: true,
        flag_end: false,
        sequence: start + 40,
        data: &data[40..],
    };
    let _ = uud.recv(chunk);
    let got = uud.get();
    assert!(
        &data[40..] == &got[..],
        "DIFF:\n {:?}\n {:?}",
        &data[40..],
        &got[..],
    );
}
#[::tracing_test::traced_test]
#[test]
fn test_stream_uud_rolling_second_case() {
    let rand = Random::new();
    let max_window: usize = 100;
    let mut uud =
        UnreliableUnorderedDatagram::with_window_size(&rand, max_window as u32);

    let mut data = Vec::with_capacity(120);
    data.resize(data.capacity(), 0);
    for i in 0..data.len() {
        data[i] = i as u8;
    }

    let start = uud.window_start.0;

    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: true,
        flag_end: false,
        sequence: start,
        data: &data[..40],
    };
    let _ = uud.recv(chunk);
    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: true,
        flag_end: false,
        sequence: start + 50,
        data: &data[50..100],
    };
    let _ = uud.recv(chunk);
    let got = uud.get();
    assert!(
        &data[..40] == &got[..],
        "DIFF:\n {:?}\n {:?}",
        &data[..40],
        &got[..],
    );
    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: true,
        flag_end: false,
        sequence: start + 40,
        data: &data[40..100],
    };
    let _ = uud.recv(chunk);
    let chunk = Chunk {
        id: stream::ID(42),
        flag_start: true,
        flag_end: false,
        sequence: start + 100,
        data: &data[100..],
    };
    let _ = uud.recv(chunk);
    let got = uud.get();
    assert!(
        &data[40..] == &got[..],
        "DIFF:\n {:?}\n {:?}",
        &data[40..],
        &got[..],
    );
}
